(*filter function on streams *)

(* nontrivial because standard definition is not syntactically guarded:*)
(*
filter f p = 
 if p (head s) 
  then cons s (filter f (tail s))
  else filter f (tail s) <-- not guarded here
*)

(*observation: if f rejects all elements of s from a given point on, then 
filter f s is undefined *)

(*solution: assume a proof that s has infinitely many elements satisfying f,
and build the filter function on that proof*)

(*will show that the filter function does not depend on its proof argument,
and that it meets the expected (coinductive) specification for filters *)

(*systematically use Knaster-Tarski theorems for coinductive predicates*)

Require Import List.
Import ListNotations.

(* Streams and their bisimulations *)
(*use coinductive records (with destructors) as an alternative to coinductive types (with constructors), to test robustness of Knaster-Tarski approach*)

CoInductive Stream{A:Type} : Type :=
  { head : A;
    tail : Stream 
  }.

(*constructors are now derived from destructors *)
Definition cons{A:Type}(a:A)(s : Stream) : Stream :=
  {|head := a; tail :=  s|}.

(* a trivial technical lemma*)
Lemma eq_head_tail{A:Type}(s1 s2:Stream(A:=A)) : head s1 = head s2 ->
                                                 tail s1 = tail s2 ->
                                                 s1 = s2.
Proof.
intros Heqhead Heqtail.
destruct s1, s2; cbn in * ; subst.
reflexivity.
Qed.

(*Bisimulation *)
CoInductive bisim{A:Type}(s1 s2: Stream (A:=A)) : Prop :=
  {
    bis_head : head s1 = head s2;
    bis_tail :  bisim (tail s1) (tail s2)
  }.

(*functional for bisimulation *)
Record _bisim{A:Type}(R: Stream(A:=A) -> Stream(A:=A) -> Prop) (s1 s2: Stream (A:=A)):=
  {
    _bis_head : head s1 = head s2;
    _bis_tail :  R (tail s1) (tail s2)
  }.                                                 

(* Knaster-Tarski theorem for bisimulations *)
(* use cofix to prove it *)
(* will then use tarski_bisim instead of cofix to prove bisimulations *)
Lemma tarski_bisim{A:Type} :forall (R: Stream -> Stream -> Prop),
    (forall s1 s2,  R s1 s2 -> _bisim R s1 s2) ->
    forall s1 s2,  R s1 s2 -> bisim(A:=A) s1 s2.
Proof.
cofix tarski_bisim.  
intros R Hco s1 s2 HR.  
generalize (Hco _ _ HR); intro Hco'.
inversion Hco'.
constructor.
 * assumption.
 * eapply tarski_bisim ; eauto.
Qed.


(* nth element *)
Fixpoint nth {A:Type}(s:Stream)(n:nat) : A :=
  match n with
    0 => head s
  | S n' => nth (tail s) n'
  end.   

(* specific constructions for filter *)
(* future *)
Inductive F{A:Type}(s:Stream)(p : A -> bool) : Type :=
|now : p (head s) = true -> F  s p  
|later : F  (tail s) p -> F s p.

(* obtain first position where p holds from F-proof *)
Fixpoint first_p{A : Type} (s : Stream) (p : A -> bool)
      (Fsp : F s p) {struct Fsp} : nat :=
  match Fsp with
  | now _ _ _ => 0
  | later _ _ Ftailsp =>
      (if p (head s)
       then 0
       else
        S (first_p  (tail s) p Ftailsp)) 
  end.

(*a property of first_p, used afterwards *)
Lemma first_p_nth{A : Type}:
  forall (s : Stream) (p : A -> bool)   (Fsp : F s p),
     p (nth s (first_p _ _ Fsp)) = true.
Proof.
intros s p Fsp.
induction Fsp; cbn in *; auto ; subst.
case_eq (p (head s)); intro Hp ; auto.
Qed.

 (* globally future *)
CoInductive GF{A:Type}(s:Stream)(p : A -> bool) : Type :=
  {
    first : F s p;
    next : GF (tail s) p
  }.

(* removing first n elements *)
Fixpoint tailn{A:Type}(s:Stream (A:=A))(n : nat) {struct n} : Stream :=
  match n with
    0 => s
   |S n' => tailn (tail s) n'
  end.  

(* GF proof for tailn *)
Fixpoint GFn{A:Type}(s:Stream (A:=A))(p: A ->bool )(GFsp : GF s p)(n : nat) {struct n} : GF (tailn s n) p:=
  match n with
    0 => GFsp
   |S n' => GFn (tail s) p (next _ _ GFsp) n'
  end. 

(* filter function *)
CoFixpoint filter {A : Type} (s : Stream) (p : A -> bool) (GFsp : GF s p) :=
 let firstp :=  first_p _ _ (first _ _ GFsp) in
  {|
    head :=  nth s firstp;
    tail := filter (tailn s (S firstp)) _ (GFn _ _ (next _ _ GFsp) firstp)
  |}.

(* testing *)
(* first test : filtering ones from 010101.... *)

Definition is0(n:nat) : bool :=
  match n with 0 => true | _  => false end.

CoFixpoint altbit : Stream(A:=nat) :=
  {| head := 0;
     tail :=
       {| head := 1; tail := altbit |}
  |}.

Program CoFixpoint GF_altbit_is0 : GF altbit is0 :=
  {|
   first := now _ _ _ ;
   next :=
     {|
       first := later _ _ (now _ _ _) ;
       next :=  GF_altbit_is0
     |}             
  |}.


Fixpoint prefix{A:Type}(n:nat)(s:Stream) {struct n} : list A :=
  match n with
    0 => []
    |S n' =>  (head s) :: (prefix n' (tail s))       
  end.

Compute prefix 20 altbit.
Compute (prefix 20 (filter altbit is0 GF_altbit_is0)).

(* second test : filter out odd numbers from 1 2 3 4 ...*)
Fixpoint even (n : nat) : bool :=
  match n with
    0 => true
  | S n' => odd n'
  end
with     
odd  (n : nat) : bool :=
  match n with
    S n'  => even n'
  | _ => false
  end.

Definition even_or_odd : forall n,  {even n = true} + {odd n = true}.
  induction n.
  * left ; reflexivity.
  * inversion IHn; subst.
    + right ; exact H.
    + left ; exact H.
Defined.

Definition even_0 : even 0 = true.
reflexivity.
Defined.

Definition even_SS : forall n, even n = true -> even (S (S n)) = true.
  intros n H.
  exact H.
Defined.

CoFixpoint nats_from (n:nat) : Stream (A:= nat) :=
  {|
    head := n;
    tail := nats_from (S n)
  |}.

Program CoFixpoint GF_nats_from_even (n:nat)(p : even n = true)
   : GF (nats_from n) even :=
   {|
   first := now _ _ _ ;
   next :=
     {|
       first := later _ _ (now _ _ _) ;
       next := GF_nats_from_even (S (S n)) (even_SS _ p)
     |}
   |}.

Compute prefix 40 (nats_from 0).
Compute prefix 20 (filter (nats_from 0) even (GF_nats_from_even 0 even_0)). 



(* showing that filter does not depend on its proof argument *)
(* first show this for first_p *)
Lemma first_p_indep_Fproof{A : Type} (s : Stream) (p : A -> bool)
  : forall  (Fsp1 Fsp2 : F s p),  first_p _ _ Fsp1 = first_p _ _ Fsp2.
Proof.
induction Fsp1; intro Fsp2.
* dependent inversion Fsp2; subst.
  + reflexivity.
  + cbn. rewrite e; reflexivity.
* dependent inversion Fsp2; subst.
  + cbn. rewrite e. reflexivity.
  + case_eq (p (head s));  intros Hps.
   - cbn; rewrite Hps; reflexivity.
   - cbn; rewrite Hps.
     rewrite IHFsp1 with f; reflexivity.
Qed.



(* then for filter, up to bisimulation *)
Lemma filter_indep_GFproof(A : Type) :
  forall (s : Stream(A:=A)) (p : A -> bool) (GFsp1 GFsp2 : GF s p),
    bisim (filter _ _  GFsp1) (filter _ _  GFsp2).
Proof.
  intros s p GFsp1 GFsp2.
  remember (filter s p GFsp1) as f1.
  remember (filter s p GFsp2) as f2.
  assert (Hco : exists (s:Stream(A:=A)) (p : A -> bool) (GFsp1 GFsp2 : GF s p),
             f1 = filter s p GFsp1 /\ f2 = filter s p GFsp2);
  [now exists s, p, GFsp1, GFsp2 |].
  eapply tarski_bisim with
      (R := fun f1 f2 =>  exists (s : Stream) (p : A -> bool) (GFsp1 GFsp2 : GF s p),
          f1 = filter s p GFsp1 /\ f2 = filter s p GFsp2); auto.
  clear. 
  intros s1 s2 (s'1 & p & GFsp1 & GFsp2 & Heq1 & Heq2) ; subst.
  constructor.
  * cbn. erewrite first_p_indep_Fproof ; eauto.
  *  dependent inversion GFsp1 ; dependent inversion GFsp2 ; cbn.
     exists  (tailn (tail s'1) (first_p s'1 p first0)), p,
     (GFn _ _  next0  (first_p s'1 p first0)),
     (GFn _ _ next1  (first_p s'1 p first0)) ; repeat split.
     +  erewrite first_p_indep_Fproof; eauto.
Qed.



(*specification: to observe that f is a p-filter of s, we need to observe
that head (f)=head(s) and (tail f) is a p-filter of (tail s) when p (head s)and that f is a p-filter of (tail s) when not (p (head s))  *)

CoInductive filter_spec{A:Type} (p : A -> bool)
            (f s : Stream (A := A)) : Prop :=
{
  ptrue :  p (head s) = true ->
                head (f) = head(s) /\ filter_spec p (tail  f) (tail s);
  pfalse :  p (head s) = false ->
                filter_spec p f (tail s)
}.       

(*functional for filter_spec *)
Record _filter_spec{A:Type}
       (R : (A -> bool) -> Stream -> Stream -> Prop)(p : A -> bool)
        (f s : Stream (A := A)) : Prop :=
{
_ptrue :  p (head s) = true -> head (f) = head(s) /\ 
                       R p (tail  f) (tail s);
_pfalse :   p (head s) = false ->
                         R p  f (tail s)
}.

(* Knaster-Tarski theorem for filter_spec*)
(* will use this instead of cofix to prove filter_spec *)
Lemma tarski_filter_spec{A:Type} :
  forall (R : (A -> bool) -> Stream -> Stream -> Prop) (p: A -> bool),
    (forall f s,  R p f s -> _filter_spec R p f s) ->
    (forall f s,  R p f s -> filter_spec p f s).
Proof.
cofix tarski_filter_spec.
intros R p Hco f s HR.
destruct (Hco _ _ HR) as (_ptrue' & _pfalse').
constructor; intros ; intuition; 
  eapply tarski_filter_spec; eauto.
Qed.


(* twp  technical lemmas *)
Lemma first_p_0{A : Type}:
  forall (s : Stream) (p : A -> bool)
         (Fsp : F s p) (GFtp  : GF (tail s) p) ,
    p (head s) = true ->
    (first_p _ _ Fsp) = 0.
Proof.
intros s p Fsp GFtp Hp.
induction Fsp; cbn in *; auto ; subst.
  rewrite Hp ; reflexivity.
Qed.

Lemma first_p_S{A : Type}:
  forall (s : Stream) (p : A -> bool)
         (Fsp : F s p) (GFtp  : GF (tail s) p) ,
    p (head s) = false ->
    (first_p _ _ Fsp) =
    S (first_p _ _ (first _ _ GFtp)).
Proof.
intros s p Fsp GFtp Hp.
induction Fsp; cbn in *; auto ; subst.
* rewrite e in Hp; discriminate.
*  rewrite Hp.
   erewrite first_p_indep_Fproof; eauto.
Qed.

(*filter function satisfies filter_spec specification*)
Theorem filter_meets_spec {A:Type} : forall (s : Stream)
                       (p : A -> bool)(GFsp : GF s p),
        filter_spec p (filter _ _ GFsp) s.
Proof.
intros s p GFsp.
remember  (filter _ _ GFsp) as s'.
apply tarski_filter_spec with
    (R := fun p s' s  =>  exists GFsp, s' = filter s p GFsp) ;
[ | now exists GFsp].
clear.
intros f s (GFs & Heq); subst.
constructor ; intros  Hp.
* split.
+ cbn.
  rewrite first_p_0 ; auto.
  inversion GFs; auto.
+ dependent inversion GFs; auto.
  dependent inversion first0.
   -  now exists next0.
   -  cbn ; rewrite Hp;  now exists next0.
* exists  (next s p GFs).
  apply trans_eq with
 (let firstp :=  first_p _ _ (first _ _ GFs ) in
  {|
    head :=  nth s firstp;
    tail := filter (tailn s (S firstp)) _ (GFn _ _ (next _ _ GFs) firstp)
  |}).
  + apply eq_head_tail ; reflexivity.
  + destruct s; cbn.
    apply eq_head_tail; auto; cbn in *.
   - dependent inversion GFs; cbn.
     remember (first_p {| head := head0; tail := s |} p first0) as f.
     destruct f.
     ** generalize (first_p_nth(A:=A)_ _ first0); intro Hc.
        rewrite <- Heqf in Hc; cbn in Hc.
        rewrite Hc in Hp ; discriminate.
     ** rewrite Heqf; auto.
        erewrite first_p_S; eauto; cbn; auto.
   -  erewrite first_p_S; eauto; cbn; auto.
Qed.



